library IEEE;  
use IEEE.std_logic_1164.all;   
use IEEE.std_logic_signed.all;   
 
use ieee.numeric_std.all;
 
entity FIR_filter is  -- VHDL projects
	 generic	(  
				  INPUT_WIDTH	: integer :=8;		-- input width defined by user  
				  OUTPUT_WIDTH	: integer :=16;	-- output width defined by user  
				  COEF_WIDTH	: integer :=8;		-- coefficient width defined by user  
				  NUM_TAPS		: integer :=17		-- filter order
				); 	
	 port (
			clk	: in	std_logic; 
			reset	: in	std_logic;	
			Din	: in	std_logic_vector(INPUT_WIDTH-1 downto 0);	-- input data 
			Dout	: out	std_logic_vector(OUTPUT_WIDTH-1 downto 0)	-- output data 
				); 
end FIR_filter;  
 
architecture behavioral of FIR_filter is  
	
	type coeficient_type is array (0 to NUM_TAPS-1) of std_logic_vector(COEF_WIDTH-1 downto 0);  
 -----------------------------------FIR filter coefficients----------------------------------------------------------------  
	constant coeficient: coeficient_type := ( x"FE",
										x"FE",
										x"03",
										x"07",
										x"FE",
										x"F5",
										x"03",
										x"29",
										x"3E",
										x"29",
										x"03",
										x"F5",
										x"FE",
										x"07",
										x"03",
										x"FE",
										x"FE"
										);                                        
 ----------------------------------------------------------------------------------------------                                     
	 type shift_reg_type is array (0 to NUM_TAPS-1) of std_logic_vector(INPUT_WIDTH-1 downto 0);  
	 signal shift_reg : shift_reg_type;  
	 type multipliers_out_type is array (0 to NUM_TAPS-1) of std_logic_vector(INPUT_WIDTH+COEF_WIDTH-1 downto 0);  
	 signal multipliers_out : multipliers_out_type;  
	 type adders_out_type is array (0 to NUM_TAPS-1) of std_logic_vector(INPUT_WIDTH+COEF_WIDTH-1 downto 0);  
	 signal adders_out: adders_out_type;
 
begin  
	--shift_reg(0)     <= Din;  
	multipliers_out(0)<= Din*coeficient(0);  
	adders_out(0)<= Din*coeficient(0);  
	
	shift_reg_proc: process (reset,clk)
	begin
		if (reset = '1') then
			shift_reg <= (others => (others => '0'));
		elsif (rising_edge(clk)) then
			for i in 0 to NUM_TAPS-2 loop
				shift_reg(i+1) <= shift_reg(i);
			end loop;
			shift_reg(0) <= Din;
		end if;
	end process shift_reg_proc;
	
	GEN_FIR: for i in 0 to NUM_TAPS-2 generate  
	begin         
		 -- filter multipliers_outiplication  
		 multipliers_out(i+1)<= shift_reg(i+1)*coeficient(i+1);   
		 -- filter combinational adders_outition  
		 adders_out(i+1)<=adders_out(i)+multipliers_out(i+1);  
	end generate GEN_FIR;  
   Dout <= adders_out(NUM_TAPS-1);
	
 end behavioral;
 
 